/* eej_calc.c
 * 
 * Copyright (C) 2006, 2007 Patrick Alken
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/*
 * This module contains routines related to computing the
 * statistics model for given parameter values. The model
 * coefficients must first be generated by the 'eej' program.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <errno.h>

#include <gsl/gsl_vector.h>
#include <gsl/gsl_math.h>

#include <ndlinear/gsl_multifit_ndlinear.h>

#include "common.h"

#include "eej_basis.h"
#include "eej_calc.h"
#include "eej_files.h"

static size_t read_coeffs(const char *filename, gsl_vector *x);

/*
 * This structure contains the information for the various
 * model coefficients we can use. The indices here correspond
 * to the EEJ_CALC_xxx values in eej_calc.h
 */
static struct
{
  const char *file_mean_coeffs;
  const char *file_stddev_coeffs;
  size_t n_mean_phi;
  size_t n_mean_t;
  size_t n_mean_s;
  size_t n_mean_euvac;
  size_t n_mean_lunar;
  size_t n_stddev_phi;
  size_t n_stddev_t;
  size_t n_stddev_s;
  size_t n_stddev_euvac;
  size_t n_stddev_lunar;
} model_info[] = {
  {
    FILE_CHAMP_MEAN_OUT,
    FILE_CHAMP_STDDEV_OUT,
    EEJ_MEAN_FIT_N_PHI_CHAMP,
    EEJ_MEAN_FIT_N_T_CHAMP,
    EEJ_MEAN_FIT_N_S_CHAMP,
    EEJ_MEAN_FIT_N_EUVAC_CHAMP,
    EEJ_MEAN_FIT_N_LUNAR_CHAMP,
    EEJ_STDDEV_FIT_N_PHI_CHAMP,
    EEJ_STDDEV_FIT_N_T_CHAMP,
    EEJ_STDDEV_FIT_N_S_CHAMP,
    EEJ_STDDEV_FIT_N_EUVAC_CHAMP,
    EEJ_STDDEV_FIT_N_LUNAR_CHAMP
  },
  {
    FILE_OER_MEAN_OUT,
    FILE_OER_STDDEV_OUT,
    EEJ_MEAN_FIT_N_PHI_OER,
    EEJ_MEAN_FIT_N_T_OER,
    EEJ_MEAN_FIT_N_S_OER,
    EEJ_MEAN_FIT_N_EUVAC_OER,
    EEJ_MEAN_FIT_N_LUNAR_OER,
    EEJ_STDDEV_FIT_N_PHI_OER,
    EEJ_STDDEV_FIT_N_T_OER,
    EEJ_STDDEV_FIT_N_S_OER,
    EEJ_STDDEV_FIT_N_EUVAC_OER,
    EEJ_STDDEV_FIT_N_LUNAR_OER
  },
  {
    FILE_SAC_MEAN_OUT,
    FILE_SAC_STDDEV_OUT,
    EEJ_MEAN_FIT_N_PHI_SAC,
    EEJ_MEAN_FIT_N_T_SAC,
    EEJ_MEAN_FIT_N_S_SAC,
    EEJ_MEAN_FIT_N_EUVAC_SAC,
    EEJ_MEAN_FIT_N_LUNAR_SAC,
    EEJ_STDDEV_FIT_N_PHI_SAC,
    EEJ_STDDEV_FIT_N_T_SAC,
    EEJ_STDDEV_FIT_N_S_SAC,
    EEJ_STDDEV_FIT_N_EUVAC_SAC,
    EEJ_STDDEV_FIT_N_LUNAR_SAC
  },
  {
    FILE_CHAMP_OER_MEAN_OUT,
    FILE_CHAMP_OER_STDDEV_OUT,
    EEJ_MEAN_FIT_N_PHI_CHAMP,
    EEJ_MEAN_FIT_N_T_CHAMP,
    EEJ_MEAN_FIT_N_S_CHAMP,
    EEJ_MEAN_FIT_N_EUVAC_CHAMP,
    EEJ_MEAN_FIT_N_LUNAR_CHAMP,
    EEJ_STDDEV_FIT_N_PHI_CHAMP,
    EEJ_STDDEV_FIT_N_T_CHAMP,
    EEJ_STDDEV_FIT_N_S_CHAMP,
    EEJ_STDDEV_FIT_N_EUVAC_CHAMP,
    EEJ_STDDEV_FIT_N_LUNAR_CHAMP
  }
};

/*
eej_calc_alloc()
  Allocate a workspace for computing the model

Inputs: model - EEJ_CALC_xxx (specifies which model to use)
                most users will want to specify EEJ_CALC_MODEL
                for the default CHAMP model
*/

eej_calc_workspace *
eej_calc_alloc(size_t model)
{
  eej_calc_workspace *w;
  size_t N_mean[EEJ_FIT_DIM];
  size_t N_stddev[EEJ_FIT_DIM];
  size_t s;

  if (model > 3)
    {
      fprintf(stderr, "eej_calc_alloc: invalid model\n");
      return 0;
    }

  w = (eej_calc_workspace *) calloc(1, sizeof(eej_calc_workspace));
  if (!w)
    return 0;

  N_mean[EEJ_IDX_PHI] = model_info[model].n_mean_phi;
  N_mean[EEJ_IDX_T] = model_info[model].n_mean_t;
  N_mean[EEJ_IDX_S] = model_info[model].n_mean_s;
  N_mean[EEJ_IDX_EUVAC] = model_info[model].n_mean_euvac;
  N_mean[EEJ_IDX_LUNAR] = model_info[model].n_mean_lunar;

  N_stddev[EEJ_IDX_PHI] = model_info[model].n_stddev_phi;
  N_stddev[EEJ_IDX_T] = model_info[model].n_stddev_t;
  N_stddev[EEJ_IDX_S] = model_info[model].n_stddev_s;
  N_stddev[EEJ_IDX_EUVAC] = model_info[model].n_stddev_euvac;
  N_stddev[EEJ_IDX_LUNAR] = model_info[model].n_stddev_lunar;

  w->u[EEJ_IDX_PHI] = &eej_basis_phi;
  w->u[EEJ_IDX_T] = &eej_basis_t;
  w->u[EEJ_IDX_S] = &eej_basis_s;
  w->u[EEJ_IDX_EUVAC] = &eej_basis_euvac;
  w->u[EEJ_IDX_LUNAR] = &eej_basis_lunar;

  s = EEJ_BASIS_CHAMP;
  if (model == EEJ_CALC_OERSTED)
    s = EEJ_BASIS_OERSTED;
  else if (model == EEJ_CALC_SAC)
    s = EEJ_BASIS_SAC;

  w->eej_basis_mean_p = eej_basis_alloc(N_mean, s, EEJ_BASIS_MEAN);
  w->eej_basis_stddev_p = eej_basis_alloc(N_stddev, s, EEJ_BASIS_STDDEV);
  if (!w->eej_basis_mean_p || !w->eej_basis_stddev_p)
    {
      eej_calc_free(w);
      return 0;
    }

  w->fit_mean_p = gsl_multifit_ndlinear_alloc(EEJ_FIT_DIM,
                                              N_mean,
                                              w->u,
                                              w->eej_basis_mean_p);
  w->fit_stddev_p = gsl_multifit_ndlinear_alloc(EEJ_FIT_DIM,
                                                N_stddev,
                                                w->u,
                                                w->eej_basis_stddev_p);
  if (!w->fit_mean_p || !w->fit_stddev_p)
    {
      eej_calc_free(w);
      return 0;
    }

  w->mean_coeffs = gsl_vector_alloc(w->fit_mean_p->n_coeffs);
  w->stddev_coeffs = gsl_vector_alloc(w->fit_stddev_p->n_coeffs);
  if (!w->mean_coeffs || !w->stddev_coeffs)
    {
      eej_calc_free(w);
      return 0;
    }

  w->x = gsl_vector_alloc(EEJ_FIT_DIM);
  if (w->x == 0)
    {
      eej_calc_free(w);
      return 0;
    }

  s = read_coeffs(model_info[model].file_mean_coeffs, w->mean_coeffs);
  if (!s)
    {
      eej_calc_free(w);
      fprintf(stderr, "eej_calc_alloc: Error opening %s\n",
              model_info[model].file_mean_coeffs);
      return 0;
    }

  s = read_coeffs(model_info[model].file_stddev_coeffs,
                  w->stddev_coeffs);
  if (!s)
    {
      eej_calc_free(w);
      fprintf(stderr, "eej_calc_alloc: Error opening %s\n",
              model_info[model].file_stddev_coeffs);
      return 0;
    }

  return (w);
} /* eej_calc_alloc() */

/*
eej_calc_free()
  Free workspace w
*/

void
eej_calc_free(eej_calc_workspace *w)
{
  if (!w)
    return;

  if (w->fit_mean_p)
    gsl_multifit_ndlinear_free(w->fit_mean_p);

  if (w->fit_stddev_p)
    gsl_multifit_ndlinear_free(w->fit_stddev_p);

  if (w->eej_basis_mean_p)
    eej_basis_free(w->eej_basis_mean_p);

  if (w->eej_basis_stddev_p)
    eej_basis_free(w->eej_basis_stddev_p);

  if (w->mean_coeffs)
    gsl_vector_free(w->mean_coeffs);

  if (w->stddev_coeffs)
    gsl_vector_free(w->stddev_coeffs);

  if (w->x)
    gsl_vector_free(w->x);

  free(w);
} /* eej_calc_free() */

/*
eej_calc_mean()
  Calculate the EEJ mean for given parameters

Inputs: phi - longitude (radians)
        t   - local time (hours)
        s   - season (day of year 0 - 365)
        e   - EUVAC value (W/m^2)
        tau - lunar local time (hours 0 to 24.833333)
        w   - workspace

Return: J(phi, t, s, e)
*/

double
eej_calc_mean(double phi, double t, double s, double e, double tau,
              eej_calc_workspace *w)
{
  double tau_r;
  double J;

  if (t < EEJ_BASIS_LT_MIN || t > EEJ_BASIS_LT_MAX)
    {
      fprintf(stderr,
              "eej_calc_mean: warning: local time (%f) outside of interval [%f, %f]\n",
              t,
              EEJ_BASIS_LT_MIN,
              EEJ_BASIS_LT_MAX);
    }

  tau_r = 2.0 * M_PI / 24.83333333 * tau;

  gsl_vector_set(w->x, EEJ_IDX_PHI, phi);
  gsl_vector_set(w->x, EEJ_IDX_T, t);
  gsl_vector_set(w->x, EEJ_IDX_S, s);
  gsl_vector_set(w->x, EEJ_IDX_EUVAC, e);
  gsl_vector_set(w->x, EEJ_IDX_LUNAR, tau_r);

  J = gsl_multifit_ndlinear_calc(w->x, w->mean_coeffs, w->fit_mean_p);

  return (J);
} /* eej_calc_mean() */

/*
eej_calc_stddev()
  Calculate the EEJ standard deviation for given parameters

Inputs: phi - longitude (radians)
        t   - local time (hours)
        s   - season (day of year 0 - 365)
        e   - EUVAC value (W/m^2)
        tau - lunar local time (lunar hours from 0 to 24.833333)
        w   - workspace

Return: sigma(phi, t, s, e, tau)
*/

double
eej_calc_stddev(double phi, double t, double s, double e, double tau,
                eej_calc_workspace *w)
{
  double tau_r;
  double sigma;

  if (t < EEJ_BASIS_LT_MIN || t > EEJ_BASIS_LT_MAX)
    {
      fprintf(stderr,
              "eej_calc_stddev: warning: local time (%f) outside of interval [%f, %f]\n",
              t,
              EEJ_BASIS_LT_MIN,
              EEJ_BASIS_LT_MAX);
    }

  tau_r = 2.0 * M_PI / 24.83333333 * tau;

  gsl_vector_set(w->x, EEJ_IDX_PHI, phi);
  gsl_vector_set(w->x, EEJ_IDX_T, t);
  gsl_vector_set(w->x, EEJ_IDX_S, s);
  gsl_vector_set(w->x, EEJ_IDX_EUVAC, e);
  gsl_vector_set(w->x, EEJ_IDX_LUNAR, tau_r);

  sigma = gsl_multifit_ndlinear_calc(w->x, w->stddev_coeffs, w->fit_stddev_p);

  return (sigma);
} /* eej_calc_stddev() */

/********************************************
 *             INTERNAL ROUTINES            *
 ********************************************/

#define MAX_BUF  1000

/*
read_coeffs()
  Read the coefficients in 'filename' and store them in 'x'
*/

static size_t
read_coeffs(const char *filename, gsl_vector *x)
{
  FILE *fp;
  size_t cnt;
  char buf[MAX_BUF];
  int n;
  double y;

  fp = fopen(filename, "r");
  if (!fp)
    {
      perror("fopen");
      return 0;
    }

  cnt = 0;
  while (fgets(buf, MAX_BUF, fp) != NULL)
    {
      if (*buf == '#')
        continue;

      n = sscanf(buf, "%le\n", &y);
      if (n <= 0)
        continue;

      gsl_vector_set(x, cnt++, y);
    }

  fclose(fp);

  return (cnt);
} /* read_coeffs() */
